///////////////////////////////////////////////////////////////////////////
// Filename : d_mem.cpp
// Project  : Data Memory Model
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision :
// Comment  : Memory Model
//      1. Synchronous Read-Write (rising-edge of in_clk, out_clk)
//      2. Data Input/Output (Dual-Port)
//      3. Write enable control input (WEN='1' for write, WEN='0' for Read)
//      4. Memory Download/Dump utility
//      5. Parametized-Reusable (MEM_SIZE/MEM_WIDTH/MEM_ADDR)
//

#include "d_mem.h"

//////////////////////////////////////////////////////////////////////////
// Memory Management Procedure
//
void d_mem::proc_address_mux()
{
    if (scl_ready.read()==(sc_bit)'0')
    {
        address = scl_address;
        data    = scl_data;
    }

    if (huff_ready.read()==(sc_bit)'0')
    {
        address = huff_address;
        data    = huff_data;
    }
}

void d_mem::proc_wen_mux()
{
    wen = scl_wen | huff_wen;
    in_clk = scl_inclk | huff_inclk;
}

//////////////////////////////////////////////////////////////////////////
// Memory Write Procedure
//
void d_mem::proc_d_mem_write()
{
    sc_uint<MEM_WIDTH>  t_data;
    sc_uint<MEM_ADDR>   t_address;

    t_data = data.read();
    t_address = address.read();

    if (((int)t_address>=MEM_SIZE) || ((int)t_address<0))   // Memory Address Range
    {
        cout << "D_MEM Write Error : Address out-of-range(" << t_address << ")" << endl;
        return;
    }

    if (wen.read()==(sc_bit)'1')    // Write Enable
    {
        //cout << "D_MEM Write :" << " Address=" << t_address << " Data=" << t_data << endl;
        ((sc_uint<MEM_WIDTH>*)pMem)[t_address] = t_data;
    }
    else
        cout << "D_MEM Write : No Write Enable" << endl;
}

//////////////////////////////////////////////////////////////////////////
// Memory Download Procedure
//
void d_mem::proc_d_mem_download()
{
    sc_uint<MEM_WIDTH>  t_data;
    sc_uint<MEM_ADDR>   t_address;

    if (download.read()==false)
        return;

    // Download File /////////////////////////////
    FILE *fp = fopen(DOWNLOAD_FILE, "r");
    int  nAddr;
    char buff[MEM_WIDTH*2], vector[MEM_WIDTH*2];

    if (!fp)
    {
        cout << "D_MEM Download Error : File (" << DOWNLOAD_FILE << ") Not Found" << endl;
        return;
    }
    else
    {
        while(fgets(buff, (MEM_WIDTH*2-1), fp))
        {
            if (!strlen(buff) || buff[0]=='\n')
            {
                cout << "D_MEM Download End :" << endl;
                fclose(fp);
                return;
            }

            if (buff[0]!='#')   // Comment Start with '#'
            {
                nAddr = get_addr(buff);
                if (nAddr<0 || nAddr>=MEM_SIZE) // Check Valid Address
                {
                    cout << "D_MEM Write Error : Address out-of-range(" << t_address << ")" << endl;
                    fclose(fp);
                    return;
                }
                t_address = nAddr;

                if (get_data(buff, vector)==0)  // Check Valid Data
                {
                    cout << "D_MEM Download Error : No Data" << endl;
                    fclose(fp);
                    return;
                }

                if (strlen(vector)!=MEM_WIDTH)      // Valid Data With
                {
                    cout << "D_MEM Download Error : Invalid Data Format" << endl;
                    fclose(fp);
                    return;
                }
                t_data = (sc_bv<MEM_WIDTH>)vector;
                ((sc_uint<MEM_WIDTH>*)pMem)[t_address] = t_data;    // Write into Memory
            }
        }

        cout << "D_MEM Download End :" << endl;
    }
    fclose(fp);
}

//////////////////////////////////////////////////////////////////////////
// Memory Read Procedure
//
void d_mem::proc_d_mem_read()
{
    static sc_uint<MEM_WIDTH>   old_data = 0;
    sc_uint<MEM_ADDR>   t_address;

    t_address = address.read();

    if (((int)t_address>=MEM_SIZE) || ((int)t_address<0))   // Memory Address Range
    {
        cout << "D_MEM Read Error : Address out-of-range(" << t_address << ")" << endl;
        return;
    }

    if (wen.read()==(sc_bit)'0')    // Memory Read
    {
        old_data = ((sc_uint<MEM_WIDTH>*)pMem)[t_address];
        //cout << "D_MEM Read :" << " Address=" << t_address << " Data=" << old_data << endl;
    }

    t_q = (int)old_data;
    q = (int)old_data;
}

//////////////////////////////////////////////////////////////////////////
// Memory Dump Procedure
//
void d_mem::proc_d_mem_dump()
{
    sc_uint<MEM_WIDTH>  t_data;
    unsigned    t_address;

    if (dump.read()==false)
        return;

    // Dump File /////////////////////////
    FILE *fp = fopen(DUMP_FILE, "w");
    char vector[MEM_WIDTH*2];

    if (!fp)
    {
        cout << "D_MEM Dump Error : File (" << DUMP_FILE << ") Open" << endl;
        return;
    }
    else
    {
        for (t_address=0; t_address<MEM_SIZE; t_address++)
        {
            t_data = ((sc_uint<MEM_WIDTH>*)pMem)[t_address];
            for (int i=0; i<MEM_WIDTH; i++)
                vector[MEM_WIDTH-i-1] = '0' + t_data[i];    // Convert to ASCII '0' or '1'
            vector[MEM_WIDTH] = '\0';

            fprintf(fp, "%03d %s\n", t_address, vector);
        }
    }
    cout << "D_MEM Dump End :" << endl;

    fclose(fp);
}

//////////////////////////////////////////////////////////////////
// Get Address
int d_mem::get_addr(char *buff)
{
    int nStr = strlen(buff);
    int i, nAddr;
    char c;

    if (nStr==0) return(-1);

    i = 0;
    while((buff[i]==' ' || buff[i]=='\t') && i < nStr) i++; // Skip Null space
    if (i>= nStr)
        return(-1);
    else
    {
        strcpy(buff, (buff+i));
        i = 0;
    }

    while((buff[i]>='0' && buff[i]<='9') && i < nStr) i++;  // Skip Null space
    c = buff[i];
    buff[i] = '\0';
    nAddr = atoi(buff);
    buff[i] = c;

    strcpy(buff, (buff+i));

    return(nAddr);
}

//////////////////////////////////////////////////////////////////
// Get Data
char* d_mem::get_data(char *buff, char* vector)
{
    int nStr = strlen(buff);
    int i, j;

    if (nStr==0) return(0);

    i = 0;
    while((buff[i]==' ' || buff[i]=='\t') && i < nStr) i++; // Skip Null space
    if (i>= nStr)
        return(0);
    else
    {
        strcpy(buff, (buff+i));
        i = 0;
    }

    j = 0;
    while((buff[i]=='0' || buff[i]=='1' || buff[i]=='X' || buff[i]=='Z') && i < nStr)
        vector[j++] = buff[i++];

    vector[j] = '\0';
    strcpy(buff, (buff+i));

    return(vector);
}
