///////////////////////////////////////////////////////////////////////////
// Filename : Mp3Stream.cxx
// Project  : Ans_MP3_Player/MP3_Decoder
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision : 5-4-01: Modified for interface to untimed SystemC testbench
//                    - John Stickley, Ikos Systems, Inc. (johns@ikos.com)
// Comment  :
//
//
 
#include "Mp3Stream.hxx"

//===========================================================================
// Mp3Stream                                                  johnS 4-21-01
//===========================================================================

// Bit Rate : 1: MPEG-1, 0: MPEG-2 LSF
static int bitrate[2][3][15] = {
    {   {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256},
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
        {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160} },
    {   {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448},
        {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
        {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320} }
    };
 
// Sampling Frequency : 1: MPEG-1, 0: MPEG-2 LSF
static double sampFreq[2][4] = {
    {22.05, 24, 16, 0}, /* MPEG-2 LSF */
    {44.1, 48, 32, 0}   /* MPEG-1 */
    };

static char *strModeExt[4] = {
    "", "[Intensity]", "[MS]", "[Intensity-MS]" };

static char *strStereo[4]   = {
    "Stereo", "J-Stereo", "Dual Ch", "Single Ch" };


//---------------------------------------------------------
// ::requestFrameInThread() - Fill next buffer from input
//                            stream on demand
//
// Description                                johnS 4-21-01
// -----------
// This is a 'slave' thread that services a request
// the frame buffer to pre-fill the next buffer from the
// input stream in preparation for laster transmission of
// header and data.  This method also checks for end-of-stream
// and notifies notifies the play control module if this is the
// case.
//---------------------------------------------------------

void Mp3Stream::requestFrameInThread(){
    if( fillStreamBuffer() == 0 ){
        fclose( dIfp );
        dIfp = NULL;
        EosOut();
        }
}

//---------------------------------------------------------
// ::frameHeaderOutThread() - Provide MP3 frame header and data
//                            to frame buffer on demand.
//
// Description                                johnS 4-21-01
// -----------
// This is a 'slave' thread that services a request
// the frame buffer for the next MP3 frame header
//---------------------------------------------------------

void Mp3Stream::frameHeaderOutThread(){

    // Find Frame's SYNC from Input Bitstream
    if( !(dMp3Header.FrameHeader=seekHeader()) )
        cout << ">>>>> SYNC Error : Frame header cannot be located <<<<<"
            << endl;
    else {
        displayToConsole();

        // CRC Error Protection
        if( Mp3HeaderProtection(dMp3Header.FrameHeader) ){
            // Purge CRC - not used here.
            getBits(16);
            }

        // Get Layer III Side Information
        getSideHeaderBlock(&dMp3Header.SideHeaderBlock);

        // Compute slot count in main data
        int nSlots = getMainDataSlots();

        dOutBuffer.WriteIndex = 0;
        for (; nSlots > 0; nSlots--)        // Read main data.
            putOutBuffer( getBits(8) );
        FrameHeaderOut = &dMp3Header;
        }
}

//---------------------------------------------------------
// ::fillStreamBuffer() - Fill buffer from MP3 stream
//
// Description                                johnS 4-21-01
// -----------
// This routine replenishes empty space in the circular
// stream buffer with more data from the FILE stream.
//
// Return != 0 if more data, == 0 if EOS
//
// Attention: This routine assumes that between each of the
// times that it is called, at least 1 byte of the written
// data is consumed (read).  If this is not the case, it
// will overwrite data that is never read.
//---------------------------------------------------------

int Mp3Stream::fillStreamBuffer() {
    unsigned bytesRead;

    if( dInBuffer.WriteIndex >= dInBuffer.ReadIndex ){
        if( (bytesRead=fread(
                &dInBuffer.Buffer[dInBuffer.WriteIndex],
                sizeof(unsigned char), FRAME_SIZE-dInBuffer.WriteIndex,
                dIfp )) != FRAME_SIZE-dInBuffer.WriteIndex ){
            dInBuffer.WriteIndex += bytesRead;
            return bytesRead; // =0: EOS, !=0 more data
            }
        dInBuffer.WriteIndex
            = (dInBuffer.WriteIndex + bytesRead) % FRAME_SIZE;

        if( dInBuffer.ReadIndex ){
            if( (bytesRead=fread(
                    &dInBuffer.Buffer[dInBuffer.WriteIndex],
                    sizeof(unsigned char), dInBuffer.ReadIndex,
                    dIfp )) != dInBuffer.ReadIndex ){
                dInBuffer.WriteIndex
                    = (dInBuffer.WriteIndex + bytesRead) % FRAME_SIZE;
                return bytesRead; // =0: EOS, !=0 more data
                }
            }
        else bytesRead = 0;
        }
    else {
        unsigned readSize = dInBuffer.ReadIndex - dInBuffer.WriteIndex;

        if( (bytesRead=fread(
                &dInBuffer.Buffer[dInBuffer.WriteIndex],
                sizeof(unsigned char), readSize, dIfp )) != readSize ){
            dInBuffer.WriteIndex
                = (dInBuffer.WriteIndex + bytesRead) % FRAME_SIZE;
            return bytesRead; // =0: EOS, !=0 more data
            }
        }
    dInBuffer.WriteIndex = (dInBuffer.WriteIndex + bytesRead) % FRAME_SIZE;
    assert( dInBuffer.WriteIndex == dInBuffer.ReadIndex );
    return 1; // =0: EOS, !=0 more data
}

//---------------------------------------------------------
// ::seekHeader() - Seek MP3 frame header in input stream
//
// Description                                johnS 4-21-01
// -----------
// This routine looks for sync bits (0xfff) in the stream
// which mark the begining of a new frame.  See Mp3Common.h
// for a description of the MP3 frame header word.
//---------------------------------------------------------

// Seek MP3 Frame Header
Mp3FrameHeader Mp3Stream::seekHeader() {
    int aligning;
    const unsigned syncPattern = 0x0FFF;  /* SYNC : 4095 = "1111_1111_1111" */
 
    aligning = dInBufferTotalBits % 8;    // BYTE Aligning
    if( aligning )
        getBits( 8-aligning );
 
    Mp3FrameHeader headerWord = getBits(32);
 
    while( (( Mp3HeaderSync(headerWord)&syncPattern) != syncPattern)
            && dInBuffer.ReadIndex != dInBuffer.WriteIndex ){
        fprintf(stderr, "Seek Sync again...\n");
        headerWord <<= 8;
        headerWord  |= getBits(8);
        }
 
    return dInBuffer.ReadIndex == dInBuffer.WriteIndex ? 0 : headerWord;
}

//---------------------------------------------------------
// ::getSideHeaderBlock(Mp3SideHeaderBlock *shb)
//     - Get side header info from MP3 frame header
//
// Description                                johnS 4-21-01
// -----------
//
//  Get Side Information of Layer III
//  ---------------------------------
//  Read following ......
//
//  - main_data_begin
//  - private_bits
//  - scfsi[ch][scfsi_band]
//  - part2_3_length[gr][ch]
//  - big_value[gr][ch]
//  - global_gain[gr][ch]
//  - scalefac_compress[gr][ch]
//  - window_switching_flag[gr][ch]
//
//  if (window_switching_flag)
//      - block_type[gr][ch]
//      - mixed_block_flag[gr][ch]
//      - table_select[gr][ch][region]
//      - subblock_gain[gr][ch][window]
//        n_region=2, n_window=3
//  else
//      - table_select[gr][ch][region]
//      - region0_count[gr][ch]
//      - region1_count[gr][ch]
//        n_region=3
//
//  - preflag[gr][ch]
//  - scalefac_scale[gr][ch]
//  - count1table_select[gr][ch]
//
//  n_ch=2 for Stereo, n_ch=1 for Mono
//  n_gr=1 for MPEG-2, n_gr=2 for MPEG-1
//---------------------------------------------------------

void Mp3Stream::getSideHeaderBlock( Mp3SideHeaderBlock *shb ){
    int ch, gr, i;
    int stereo = Mp3HeaderStereo(dMp3Header.FrameHeader) + 1;
    int version = Mp3HeaderVersion(dMp3Header.FrameHeader);
   
    if( version != MPEG_PHASE2_LSF ) { /* MPEG-1 */
        shb->MainDataBegin = getBits(9);
        if (stereo == 1)
            shb->PrivateBits = getBits(5);
        else
            shb->PrivateBits = getBits(3);

        for( ch=0; ch<stereo; ch++ )
            shb->Ch[ch].ScfsiBits = getBits(4);
 
        for( gr=0; gr<2; gr++ )
            for( ch=0; ch<stereo; ch++ ) {
                shb->Ch[ch].Gr[gr].Part2_3Length       = getBits(12);
                shb->Ch[ch].Gr[gr].BigValues           = getBits(9);
                shb->Ch[ch].Gr[gr].GlobalGain          = getBits(8);
                shb->Ch[ch].Gr[gr].ScalefacCompress    = getBits(4);
                shb->Ch[ch].Gr[gr].WindowSwitchingFlag = getBits(1);

                if( shb->Ch[ch].Gr[gr].WindowSwitchingFlag ) {
                    shb->Ch[ch].Gr[gr].BlockType           = getBits(2);
                    shb->Ch[ch].Gr[gr].MixedBlockFlag      = getBits(1);
                    for (i=0; i<2; i++)
                        shb->Ch[ch].Gr[gr].TableSelect[i]  = getBits(5);
                    for (i=0; i<3; i++)
                        shb->Ch[ch].Gr[gr].SubblockGain[i] = getBits(3);

                    /* Set region_count parameters since they are
                       implicit in this case. */
                    if( shb->Ch[ch].Gr[gr].BlockType == 0 )
                        errorSideInfo();
                    else if( shb->Ch[ch].Gr[gr].BlockType == 2
                            && shb->Ch[ch].Gr[gr].MixedBlockFlag == 0 )
                        shb->Ch[ch].Gr[gr].Region0Count = 8; /* MI 9; */
                    else
                        shb->Ch[ch].Gr[gr].Region0Count = 7; /* MI 8; */
                    shb->Ch[ch].Gr[gr].Region1Count
                        = 20 - shb->Ch[ch].Gr[gr].Region0Count;
                    }
                else {
                    for( i=0; i<3; i++ )
                        shb->Ch[ch].Gr[gr].TableSelect[i] = getBits(5);
                    shb->Ch[ch].Gr[gr].Region0Count       = getBits(4);
                    shb->Ch[ch].Gr[gr].Region1Count       = getBits(3);
                    shb->Ch[ch].Gr[gr].BlockType          = 0;
                    }

                shb->Ch[ch].Gr[gr].Preflag           = getBits(1);
                shb->Ch[ch].Gr[gr].ScalefacScale     = getBits(1);
                shb->Ch[ch].Gr[gr].Count1tableSelect = getBits(1);
                }
        }
    else { /* Layer 3 LSF */
        shb->MainDataBegin   = getBits(8);
        if( stereo == 1 )
            shb->PrivateBits = getBits(1);
        else
            shb->PrivateBits = getBits(2);

        gr=0;

        for( ch=0; ch<stereo; ch++ ) {
            shb->Ch[ch].Gr[gr].Part2_3Length       = getBits(12);
            shb->Ch[ch].Gr[gr].BigValues           = getBits(9);
            shb->Ch[ch].Gr[gr].GlobalGain          = getBits(8);
            shb->Ch[ch].Gr[gr].ScalefacCompress    = getBits(9);
            shb->Ch[ch].Gr[gr].WindowSwitchingFlag = getBits(1);
            
            if( shb->Ch[ch].Gr[gr].WindowSwitchingFlag ) {
                shb->Ch[ch].Gr[gr].BlockType      = getBits(2);
                shb->Ch[ch].Gr[gr].MixedBlockFlag = getBits(1);
                for( i=0; i<2; i++ )
                    shb->Ch[ch].Gr[gr].TableSelect[i] = getBits(5);
                for( i=0; i<3; i++ )
                    shb->Ch[ch].Gr[gr].SubblockGain[i] = getBits(3);
                
                /* Set region count parameters since they are implicit 
                   in this case. */

                if( shb->Ch[ch].Gr[gr].BlockType == 0 )
                    errorSideInfo();
                else if( shb->Ch[ch].Gr[gr].BlockType == 2
                        && shb->Ch[ch].Gr[gr].MixedBlockFlag == 0 )
                    shb->Ch[ch].Gr[gr].Region0Count = 8; /* MI 9; */
                else
                    shb->Ch[ch].Gr[gr].Region0Count = 7; /* MI 8; */

                shb->Ch[ch].Gr[gr].Region1Count
                    = 20 - shb->Ch[ch].Gr[gr].Region0Count;
                }
            else {
                for( i=0; i<3; i++ )
                    shb->Ch[ch].Gr[gr].TableSelect[i] = getBits(5);
                shb->Ch[ch].Gr[gr].Region0Count       = getBits(4);
                shb->Ch[ch].Gr[gr].Region1Count       = getBits(3);
                shb->Ch[ch].Gr[gr].BlockType          = 0;
                }
            
            shb->Ch[ch].Gr[gr].ScalefacScale     = getBits(1);
            shb->Ch[ch].Gr[gr].Count1tableSelect = getBits(1);
            }   /* for Channel */
        }   /* Layer 3 */
}

//---------------------------------------------------------
// ::getMainDataSlots() - Get number of main data slots
//
// Description                                johnS 4-21-01
// -----------
//---------------------------------------------------------

int Mp3Stream::getMainDataSlots() {
    int nSlots;

    int stereo = Mp3HeaderStereo(dMp3Header.FrameHeader) + 1;
 
    /* Frame Size */
    nSlots = (int)(
        (144 * bitrate
            [Mp3HeaderVersion(dMp3Header.FrameHeader)]
            [2]
            [Mp3HeaderBitrateID(dMp3Header.FrameHeader)])
        / sampFreq
            [Mp3HeaderVersion(dMp3Header.FrameHeader)]
            [Mp3HeaderSamplingFrequency(dMp3Header.FrameHeader)]
        );
 
    /* Side_Info Size */
    if ( Mp3HeaderVersion(dMp3Header.FrameHeader) != MPEG_PHASE2_LSF ) {   /* MPEG-1 */
        if( stereo == 1 )
            nSlots -= 17;
        else
            nSlots -= 32;
        }
    else {   /* MPEG-2 LSF */
        nSlots = nSlots / 2;
        if( stereo == 1 )
            nSlots -= 9;
        else
            nSlots -=17;
        }
 
    if( Mp3HeaderPadding(dMp3Header.FrameHeader) ) // Account for padding.
        nSlots++;
 
    nSlots -= 4; // Deduct for size of frame header.
 
    if( Mp3HeaderProtection(dMp3Header.FrameHeader) )
        nSlots -= 2; // Deduct for CRC size
 
    return(nSlots);
}

//---------------------------------------------------------
// ::displayToConsole() - Display progress data to console
//
// Description                                johnS 4-21-01
// -----------
// This routine displays MP3 header info and a "rotating
// slash" and frame count indicators to the console.
// Additionally, it does a one-time send of information
// to the PcmStream module for use in creating the output
// file.
//---------------------------------------------------------

void Mp3Stream::displayToConsole() {
    if( dFirstConsoleDisplay ){
        dFirstConsoleDisplay = 0;
 
        int version   = Mp3HeaderVersion(dMp3Header.FrameHeader);
        int bitrateID = Mp3HeaderBitrateID(dMp3Header.FrameHeader);
        int mode      = Mp3HeaderMode(dMp3Header.FrameHeader);
        int modeExt   = Mp3HeaderModeExt(dMp3Header.FrameHeader);
        int stereo    = Mp3HeaderStereo(dMp3Header.FrameHeader);
        double samplingFrequency
            = sampFreq
                [version]
                [Mp3HeaderSamplingFrequency(dMp3Header.FrameHeader)];

        // Header Info ////////////////////////////
        cout << "MPEG-" << (version ? 1 : 2) << endl;
        cout << "Bitrate: " << bitrate[version][2][bitrateID] << " Kbps" << endl;
        cout << "Sampling Frequency: " << samplingFrequency << " Khz" << endl;
        cout << "Mode(Stereo): "
             << strStereo[mode] << strModeExt[modeExt] << endl;
 
        PcmFileInfo pcmFileInfo;

        pcmFileInfo.Stereo = stereo;
        pcmFileInfo.SamplingFrequency = (int)(samplingFrequency*1000);

        char temp[32];
        dPcmFileName += stereo ? ".stereo" : ".mono";
        sprintf( temp, ".%d", (int)(samplingFrequency*1000) );
        dPcmFileName += temp;
        dPcmFileName += ".wav";

        pcmFileInfo.FileName = dPcmFileName.c_str();
        FileHeaderOut = &pcmFileInfo;
        }
 
    // Display frame counts, rotating slash '/'
    static int frameCount = 0;

    frameCount++;
    if(frameCount % 100 == 0)
        fprintf(stderr, "\015{%d}\n", frameCount);
    else {
        if( (frameCount&0xf) == 0 ){
            switch( ((frameCount>>4)&0x3) ) {
                case 0: fprintf(stderr, "\015-");  break;
                case 1: fprintf(stderr, "\015\\"); break;
                case 2: fprintf(stderr, "\015|");  break;
                case 3: fprintf(stderr, "\015/");  break;
                }
            fflush(stderr);
            }
        }
}
