///////////////////////////////////////////////////////////////////////////
// Filename : sc_III_Huffman.cpp
// Project  : Ans_MP3_Player/MP3_Decoder
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision : 5-4-01: Modified for interface to untimed SystemC testbench
//                    - John Stickley, Ikos Systems, Inc. (johns@ikos.com)
// Comment  :
//
//

#include "FrameBuffer.hxx"
#include "sc_III_Decoder_Mem.h"
#include "sc_III_Huffman.h"

#include "sfBandIndex.h"

void iii_huffman::proc_huffman() {

    if (rst.read()==(sc_bit)'0')    // Async. Reset
    {
        Huffman_State = S_HUFFMAN_WAIT;
        ready = (sc_bit)'1';
    }
    else if (clk.event() && clk.read()==(sc_bit)'1')    // Clocking Rising-Edge
    {
        switch(Huffman_State.read())
        {
            case S_HUFFMAN_WAIT:
                if (start.read()==(sc_bit)'1')
                {
                    ready = (sc_bit)'0';
                    Huffman_State = S_HUFFMAN_DECODE;
                }
                break;

            case S_HUFFMAN_DECODE:
                if (start.read()==(sc_bit)'0') {
                    Mp3FrameHeader frameHeader      = FrameHeader.read();
                    const Mp3SideHeader *sideHeader = SideHeader.read();

                    int t_ch = ch.read();
                    int t_gr = gr.read();
                    unsigned long t_part2_start = part2_start.read();

                    III_hufman_decode(
                        Mp3HeaderVersion(frameHeader),
                        sideHeader->BigValues,
                        sideHeader->TableSelect[0],
                        sideHeader->TableSelect[1],
                        sideHeader->TableSelect[2],
                        sideHeader->Count1tableSelect,
                        Mp3HeaderSamplingFrequency(frameHeader),
                        sideHeader->Region0Count,
                        sideHeader->Region1Count,
                        sideHeader->Part2_3Length,
                        sideHeader->WindowSwitchingFlag,
                        sideHeader->BlockType,
                        t_ch, t_gr,
                        t_part2_start);

                    Huffman_State = S_HUFFMAN_DOWNLOAD;
                }
                break;

            case S_HUFFMAN_DOWNLOAD:
                Huffman_State = S_HUFFMAN_BUSY;
                break;

            case S_HUFFMAN_BUSY:
                ready = (sc_bit)'1';
                Huffman_State = S_HUFFMAN_WAIT;
                break;

            default:
                Huffman_State = S_HUFFMAN_WAIT;
                break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////
// Huffman Decoder for Layer III
//
//int huffman_initialized = FALSE;
#include <stdio.h>  // needed for definition of NULL
#include "sc_III_Huffman_table.h"

void iii_huffman::III_hufman_decode(
    int         version,
    int         big_values,
    int         table_select0,
    int         table_select1,
    int         table_select2,
    int         count1table_select,
    int         sampling_frequency,
    int         region0_count,
    int         region1_count,
    int         part2_3_length,
    int         window_switching_flag,
    int         block_type,
    int         ch,
    int         gr,
    int         part2_start)
{
    int i, x, y;
    int v, w;
    int region1Start;
    int region2Start;
    int sfreq;
    int currentBit, grBits;

    struct huffcodetab  *h;

    sfreq = sampling_frequency + (version * 3);

    // Find Region Boundary
    if (window_switching_flag && block_type==2) // Short Block
    {
        region1Start = 36;  /* sfb[9/3]*3=36 */
        region2Start = 576; /* No Region2 for short block case. */
    }
    else    // LONG Block
    {
        region1Start = sfBandIndex[sfreq].l[region0_count + 1];
        region2Start = sfBandIndex[sfreq].l[region0_count + region1_count + 2];
    }

    grBits     = part2_start + part2_3_length;
    currentBit = GlobalFrameBuffer->TellBits();
    
    // Read bigvalues area.
    for (i=0; i<(int)(big_values*2); i+=2) {
        if (i<region1Start)
            h = &ht[table_select0];
        else if (i<region2Start)
            h = &ht[table_select1];
        else
            h = &ht[table_select2];

        huffman_decoder(h, &x, &y, &v, &w);

        GlobalDecoderMem->put_is( x, ch, i/SSLIMIT, i%SSLIMIT);
        GlobalDecoderMem->put_is( y, ch, (i+1)/SSLIMIT, (i+1)%SSLIMIT);
    }

    grBits     = part2_start + part2_3_length;
    currentBit = GlobalFrameBuffer->TellBits();

    /* Read count1 area. */
    h = &ht[count1table_select+32];
    while (( (currentBit=GlobalFrameBuffer->TellBits()) < part2_start + part2_3_length ) && ( i < SSLIMIT*SBLIMIT ))
    {
        huffman_decoder(h, &x, &y, &v, &w);

        GlobalDecoderMem->put_is( v, ch, i/SSLIMIT, i%SSLIMIT);
        GlobalDecoderMem->put_is( w, ch, (i+1)/SSLIMIT, (i+1)%SSLIMIT);
        GlobalDecoderMem->put_is( x, ch, (i+2)/SSLIMIT, (i+2)%SSLIMIT);
        GlobalDecoderMem->put_is( y, ch, (i+3)/SSLIMIT, (i+3)%SSLIMIT);
        i += 4;
    }

    grBits     = part2_start + part2_3_length;
    currentBit = GlobalFrameBuffer->TellBits();

    if (currentBit > part2_start + part2_3_length) {
        i -=4;
        GlobalFrameBuffer->RewindBits(
            currentBit - part2_start - part2_3_length );
    }
    
    /* Dismiss stuffing Bits */
    grBits     = part2_start + part2_3_length;
    currentBit = GlobalFrameBuffer->TellBits();
    
    if ( currentBit < grBits )
        getBits( grBits - currentBit );
    
    /* Zero out rest. */
    for (; i<SSLIMIT*SBLIMIT; i++)
        GlobalDecoderMem->put_is( 0, ch, i/SSLIMIT, i%SSLIMIT);
}

///////////////////////////////////////////////////////////////////////////
// Huffman decoder for Layer III
//   do the huffman-decoding
//   note! for counta,countb -the 4 bit value is returned in y, discard x
//
int iii_huffman::huffman_decoder(
    struct huffcodetab  *h,
    int                 *x,
    int                 *y,
    int                 *v,
    int                 *w)
{
    HUFFBITS    level = 1 << ((sizeof(HUFFBITS) * 8) - 1);
    int         point = 0;
    int         error = 1;

    if (h->val == NULL)
        return(2);

    /* table 0 needs no bits */
    if ( h->treelen == 0)
    {
        *x = *y = 0;
        return(0);
    }

    /* Lookup in Huffman table. */
    do
    {
        if (h->val[point][0]==0)   /*end of tree*/
        {
            *x = h->val[point][1] >> 4;
            *y = h->val[point][1] & 0xf;

            error = 0;
            break;
        }

        if (getBit()) {
            while (h->val[point][1] >= MXOFF)
                point += h->val[point][1];
            point += h->val[point][1];
        }
        else {
            while (h->val[point][0] >= MXOFF)
                point += h->val[point][0];
            point += h->val[point][0];
        }

        level >>= 1;
    } while (level  || (point < (int)ht->treelen) );

    /* Process sign encodings for quadruples tables. */
    if (h->tablename[0] == '3' &&
        (h->tablename[1] == '2' || h->tablename[1] == '3'))
    {
        *v = (*y>>3) & 1;
        *w = (*y>>2) & 1;
        *x = (*y>>1) & 1;
        *y = *y & 1;

        /* v, w, x and y are reversed in the bitstream.
            switch them around to make test bistream work. */

        if (*v)
            if (getBit())
                *v = -*v;

        if (*w)
            if (getBit())
                *w = -*w;

        if (*x)
            if (getBit())
                *x = -*x;

        if (*y)
            if (getBit())
                *y = -*y;
    }
    else
    {
        // Process sign and escape encodings for dual tables.

        // x and y are reversed in the test bitstream.
        //  Reverse x and y here to make test bitstream work.

        if (h->linbits)
            if ((int)(h->xlen-1) == *x)
                *x += getBits(h->linbits);
        if (*x)
            if (getBit())
                *x = -*x;

        if (h->linbits)
            if ((int)(h->ylen-1) == *y)
                *y += getBits(h->linbits);

        if (*y)
            if (getBit())
                *y = -*y;
    }

    return(error);
}


