///////////////////////////////////////////////////////////////////////////
// Filename : sc_III_Hybrid.cpp
// Project  : Ans_MP3_Player/MP3_Decoder
// Author   : GoodKook / goodkook@nms.anslab.co.kr
// Company  : AnsLab Co. / http://www.anslab.co.kr
// Revision : 5-4-01: Modified for interface to untimed SystemC testbench
//                    - John Stickley, Ikos Systems, Inc. (johns@ikos.com)
// Comment  :
//
//

#include "sc_III_Hybrid.h"

#include "sc_III_Decoder_Mem.h"

void iii_hybrid::proc_hybrid()
{

    if (rst.read()==(sc_bit)'0')    // Async. Reset
    {
        Hybrid_State = S_HYBRID_WAIT;
        ready = (sc_bit)'1';
    }
    else if (clk.event() && clk.read()==(sc_bit)'1')    // Clocking Rising-Edge
    {
        switch(Hybrid_State.read())
        {
            case S_HYBRID_WAIT:
                if (start.read()==(sc_bit)'1')
                {
                    ready = (sc_bit)'0';
                    Hybrid_State = S_HYBRID_DECODE;
                }
                break;

            case S_HYBRID_DECODE:
                if (start.read()==(sc_bit)'0') {
                    const Mp3SideHeader *sideHeader = SideHeader.read();
                    int     t_ch                    = channel.read();
                    III_hybrid(
                        sideHeader->WindowSwitchingFlag,
                        sideHeader->MixedBlockFlag,
                        sideHeader->BlockType,
                        t_ch );

                    Hybrid_State = S_HYBRID_DOWNLOAD;
                }
                break;

            case S_HYBRID_DOWNLOAD:
                Hybrid_State = S_HYBRID_BUSY;
                break;

            case S_HYBRID_BUSY:
                ready = (sc_bit)'1';
                Hybrid_State = S_HYBRID_WAIT;
                break;

            default:
                Hybrid_State = S_HYBRID_WAIT;
                break;
        }
    }
}

/////////////////////////////////////////////////////////////////////////
// MPEG Layer III Hybrid Filter Bank
//  1. IMDCT
//  2. Overlapped Addition
//  3. Frequency Inversion
//
void iii_hybrid::III_hybrid(
    int     window_switching_flag,
    int     mixed_block_flag,
    int     block_type,
    int     ch )
{
    int     sb, ss;
    int     bt;
    double  temp;

    for (sb=0; sb<SBLIMIT; sb++)    // Hybrid synthesis.
    {
        bt = (window_switching_flag && mixed_block_flag && (sb < 2)) ? 0 : block_type; 

        // IMDCT
        inv_mdct(ch, sb, bt);

        for(ss=0; ss<SSLIMIT; ss++)
        {
            // Overlap addition
            temp = GlobalDecoderMem->get_rawout(ss)
                    + GlobalDecoderMem->get_prevblck(ch, sb, ss);
            GlobalDecoderMem->put_hybridOut(temp , ch, sb, ss);

            // Save last half
            temp = GlobalDecoderMem->get_rawout(ss+18);
            GlobalDecoderMem->put_prevblck(temp, ch, sb, ss);

            // Frequency inversion for polyphase
            if ((ss%2) && (sb%2))
            {
                temp = -(GlobalDecoderMem->get_hybridOut(ch, sb, ss));
                GlobalDecoderMem->put_hybridOut(temp, ch, sb, ss);
            }
        }
    }     
}


/////////////////////////////////////////////////////////////////////////
// MPEG Layer III Hybrid Filter Bank (IMDCT)
//
void iii_hybrid::inv_mdct(int ch, int sb, int block_type)
{
    int     i,m,N,p;
    double  tmp[12],sum;

    for(i=0;i<36;i++)
        GlobalDecoderMem->put_rawout(0.0, i);

    if(block_type == 2) // SHORT
    {
        N=12;
        for(i=0;i<3;i++)
        {
            for(p=0;p<N;p++)
            {
                sum = 0.0;
                for(m=0;m<N/2;m++)
                    sum += ( COS12[p*6+m] * GlobalDecoderMem->get_hybridIn(ch, sb, (i+3*m)) );
                    //sum += ( cos( PI/(double)(2*N)*(double)(2*p+1+N/2)*(double)(2*m+1) ) *
                    //          GlobalDecoderMem->get_hybridIn(ch, sb, (i+3*m)) );
                tmp[p] = sum * win[block_type][p] ;
            }

            for(p=0;p<N;p++)
                GlobalDecoderMem->put_rawout(
                    (GlobalDecoderMem->get_rawout(6*i+p+6) + tmp[p]), (6*i+p+6) );
        }
    }
    else    // LONG & MIXED
    {
        N=36;
        for(p=0;p<N;p++)
        {
            sum = 0.0;
            for(m=0;m<N/2;m++)
                sum += ( COS36[((2*p+1+N/2)*(2*m+1))%(4*36)]
                            * GlobalDecoderMem->get_hybridIn(ch, sb, m));
            GlobalDecoderMem->put_rawout( (sum * win[block_type][p]), p);
        }
    }
}

void iii_hybrid::create_imdct_coeff()
{
    int i, p, m;

    // COS Coeff for Short block (12-point IMDCT)
    for(p=0;p<12;p++)
        for(m=0;m<6;m++)
            COS12[p*6+m] = cos( PI/(double)(2*12)*(double)(2*p+1+12/2)*(double)(2*m+1) );

    // COS Coeff for Short block (36-point IMDCT)
    for (i=0; i<4*36; i++)
        COS36[i] = cos(PI/(2*36) * i);
        
    // Window Coeff: type 0
    for(i=0;i<36;i++)       win[0][i] = sin( PI/36 *(i+0.5) );

    // Window Coeff: type 1
    for(i=0;i<18;i++)       win[1][i] = sin( PI/36 *(i+0.5) );
    for(i=18;i<24;i++)      win[1][i] = 1.0;
    for(i=24;i<30;i++)      win[1][i] = sin( PI/12 *(i+0.5-18) );
    for(i=30;i<36;i++)      win[1][i] = 0.0;

    // Window Coeff: type 3
    for(i=0;i<6;i++)        win[3][i] = 0.0;
    for(i=6;i<12;i++)       win[3][i] = sin( PI/12 *(i+0.5-6) );
    for(i=12;i<18;i++)      win[3][i] =1.0;
    for(i=18;i<36;i++)      win[3][i] = sin( PI/36*(i+0.5) );

    // Window Coeff: type 2
    for(i=0;i<12;i++)       win[2][i] = sin( PI/12*(i+0.5) ) ;
    for(i=12;i<36;i++)      win[2][i] = 0.0 ;
}

